/* 
 * Copyright (C) 2014 Vincent Vansuyt
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.WindowManager;
import ij.plugin.PlugIn;
import ij.process.ImageProcessor;
import ij.gui.GenericDialog;
import ij.gui.ImageWindow;

/**
 * @description
 * English description :
 * Filter Lab stack image using a selective gaussian blur.
 * It compute a ponderated mean of in the neighborhood pixel. For each neighborhood 
 *  pixel 'q' of computed pixel 'p',
 * the associated weight 'q' is computed with a gaussian function and the luminance
 *  difference between luminance 'p' and luminance 'q'. If difference is zero, 
 * weight is 1. If difference is bigger weight tend towards zero.
 * Only chrominance sclices are filtered ( sclices a and b )
 *
 * French description :
 * Filtre une image-stack contenant 3 slices L,a,b en utilisant pour chaque pixel
 *  une moyenne des pixels voisins ayant la même luminance.
 * L'écart de luminance est pris en compte par une fonction gaussienne qui renvoie
 *  le poids de prise en compte du pixel dans la moyenne.
 * Si l'écart de luminance est important, le coefficient est nul
 * Seuls les slices de chrominance (a et b) sont filtrés.
  
 * @version v1.01 : 
 *        + code improvement (40% faster)
 *        + adding timing
 * @author Vincent Vansuyt
 */
public class Lab_nonlocal_mean_with_same_pixel_luminance_neighborhood_v1_01 implements PlugIn {
	/**
	 * Point d'entree du plugin
	 *
	 * @param string
	 */
	@Override
	public void run(String string) {
		IJ.log( "\\Clear" );
		IJ.log( "------- Lab non-local mean filtering - begin ------" );
		IJ.log( "Dossier log : " + System.getProperty( "user.dir" ) );
		//imp IJ.getImage();
		ImagePlus imp = WindowManager.getCurrentImage();
		if (imp == null ) {
			IJ.noImage () ;
			return ;
		}				
		
		// − doit être stack
		int nNb_slices = imp.getNSlices();
		String strM = "Nb of slices = " + nNb_slices;
		IJ.log( strM );		
		int nWidth = imp.getWidth();
		int nHeight = imp.getHeight();
		String strNom_img_src = imp.getTitle();
		strM = "Image w:" + nWidth + " x h:" + nHeight;
		IJ.log( strM );
		if ( nNb_slices != 3 ) {
			strM = "Image must contain 3 slices (L, a, b)";
			IJ.showMessage( "Warning", strM );
			IJ.log( strM );
			return ;
		}

		ImageStack imgstackIn = imp.getImageStack();
		ImageProcessor ipIn_L = imgstackIn.getProcessor( 0 + 1 );
		ImageProcessor ipIn_a = (imgstackIn.getProcessor( 1 + 1 ));
		ImageProcessor ipIn_b = (imgstackIn.getProcessor( 2 + 1 ));
		
		// Dialogue récupération des valeurs de sigma et la taille de la 
		//  fenêtre de filtre
		GenericDialog gd = new GenericDialog( "Color filtering" );
		gd.addMessage( "Non-local mean with same luminance pixel neighbourood" );
		final int ncNb_decimales = 0;
		int nWinsz_x = 20;
		int nWinsz_y = 20;
		gd.addNumericField("Windows x size (pixels) : ", nWinsz_x, ncNb_decimales );
		gd.addNumericField("Windows y size (pixels) : ", nWinsz_y, ncNb_decimales );
		double nSigma_luminance = 4;
		int nNb_decimales = 2;
		gd.addNumericField("Sigma luminance : ", nSigma_luminance, nNb_decimales );
		gd.showDialog();
		if ( gd.wasCanceled() ) {
			IJ.log("Cancel" );
			IJ.log( "------- Lab non-local mean filtering - end ------" );
			return;
		}
		nWinsz_x = (int)gd.getNextNumber();
		IJ.log("Taille des blocs (pixels) : " + nWinsz_x );
		nWinsz_y = (int)gd.getNextNumber();
		IJ.log("Taille des blocs (pixels) : " + nWinsz_y );
		nSigma_luminance = gd.getNextNumber();
		IJ.log("Sigma : " + String.format( "%.2f", nSigma_luminance ) );
		
		// Création de l'image-stack Lab résultat
		// Le canal L reste identique
		// Les canaux a et b sont filtrés avec un filtre bilateral basé sur la luminance
		long t_debut_millisec = System.currentTimeMillis();// pour le chronometrage
		final int ncNb_sclices = 3;		
		// "RGB Black" "8− bit White"
		IJ.log("strNom_img_src = " + strNom_img_src );
		String strNom_img_dest = strNom_img_src + "_Lab_nonlocal_mean" +
				                   "_win_sz_X:" + String.format("%2d", nWinsz_x ) +
				                   "_win_sz_Y:" + String.format("%2d", nWinsz_y ) +
				                   "_sigma_lum:" + String.format("%.2f", nSigma_luminance );
		IJ.log("strNom_img_dest = " + strNom_img_dest );
		ImagePlus imp_result = IJ.createImage( strNom_img_dest, "32-bit black" , nWidth , nHeight , ncNb_sclices );
		ImageStack imgstackOut = imp_result.getImageStack();
		ImageProcessor ipOut_L = imgstackOut.getProcessor( 0 + 1 );
		ImageProcessor ipOut_a = imgstackOut.getProcessor( 1 + 1 );
		ImageProcessor ipOut_b = imgstackOut.getProcessor( 2 + 1 );		
		
		double[] clLab = {0.0, 0.0, 0.0 };
		int[] clRGB = {0,0,0};		
		IJ.log( "Parsing image..." );
		int nDemi_winsz_x = nWinsz_x / 2;
		int nDemi_winsz_y = nWinsz_y / 2;
		int y_debut = 0;
		int y_fin = nHeight - 1;
		int x_debut = 0;
		int x_fin = nWidth - 1;
		int nPas = (int)Math.round( (float)( y_fin - y_debut) / 16.0 );
		for ( int y = y_debut; y <= y_fin ; ++y ) {
			if ( y % nPas == 0 ) {
				IJ.showProgress( y, nHeight );
			}
			int y_bas = y - nDemi_winsz_y;
			int y_haut = y + nDemi_winsz_y;
			int nOffset_dy_1 = Math.max( 0, ( y_debut - y_bas ) );
			int nOffset_dy_2 = Math.max( 0, ( y_haut - y_fin ) );
			double nLum_p;
			double nLum_q;
			double g, v_a, v_b;
			for ( int x = x_debut; x <= x_fin ; ++x ) {
				int x_gauche = x - nDemi_winsz_x;
				int x_droit = x + nDemi_winsz_x;
				int nOffset_dx_1 = Math.max( 0, ( x_debut - x_gauche ) );
				int nOffset_dx_2 = Math.max( 0, ( x_droit - x_fin ) );
				// Copie du canal L
				nLum_p = ipIn_L.getf( x, y );
				ipOut_L.putPixelValue( x, y, nLum_p );
				// Filtrage sur les canaux a et b						
				clLab[1] = ipIn_a.getf( x, y );
				clLab[2] = ipIn_b.getf( x, y );				
				double nSomme_a = 0.0;
				double nSomme_b = 0.0;
				double nSomme_pds = 0.0;
				final double ncDenom_sigma = 2.0 * nSigma_luminance * nSigma_luminance;
				for (int dy = -nDemi_winsz_y + nOffset_dy_1; dy <= nDemi_winsz_y - nOffset_dy_2; ++dy ) {
					int pos_y = y + dy;
					for (int dx = -nDemi_winsz_x + nOffset_dx_1; dx <= nDemi_winsz_x - nOffset_dx_2; ++dx  ) {
						int pos_x = x + dx;
						nLum_q = ipIn_L.getf( pos_x, pos_y );
						// for a and b
						g = Math.exp( -Math.pow( nLum_p - nLum_q, 2.0 )/ncDenom_sigma );
						nSomme_pds += g;
						// a						
						v_a = ipIn_a.getPixelValue( pos_x, pos_y );
						nSomme_a += g * v_a;						
						// b
						v_b = ipIn_b.getPixelValue( pos_x, pos_y );
						nSomme_b += g * v_b;
					}
					ipOut_a.putPixelValue(x, y, nSomme_a / nSomme_pds);
					ipOut_b.putPixelValue(x, y, nSomme_b / nSomme_pds);
				}
			}			
		}
		imp_result.getProcessor().resetMinAndMax();
		imp_result.show();		
		long t_fin_millisec = System.currentTimeMillis();// pour le chronometrage
		long t_duree_millisec = t_fin_millisec - t_debut_millisec;
		double t_duree_sec = t_duree_millisec / 1000.0;
		String strTps = String.format( "%.2f", t_duree_sec );
		IJ.log( "------- Lab non-local mean filtering - end - elapsed tme : " + strTps + "s. ------" );
	}// run
}// Lab_nonlocal_mean_with_same_pixel_luminance_neighborhood_v1_01